diff --git a/src/test/perl/TestLib.pm b/src/test/perl/TestLib.pm
index 4927d45..11f774e 100644
--- a/src/test/perl/TestLib.pm
+++ b/src/test/perl/TestLib.pm
@@ -35,6 +35,7 @@ our @EXPORT = qw(
 
 use Cwd;
 use File::Basename;
+use File::Path qw(rmtree);
 use File::Spec;
 use File::Temp ();
 use IPC::Run qw(run start);
@@ -107,21 +108,24 @@ $ENV{PGPORT} = int($ENV{PGPORT}) % 65536;
 # Helper functions
 #
 
+my $temp_cleanup = 1;
+my @temp_dirs = ();
 
 sub tempdir
 {
-	return File::Temp::tempdir(
+	my $path = File::Temp::tempdir(
 		'tmp_testXXXX',
 		DIR => $ENV{TESTDIR} || cwd(),
-		CLEANUP => 1);
+		CLEANUP => 0);
+	push(@temp_dirs, $path);
+	return $path;
 }
 
 sub tempdir_short
 {
-
 	# Use a separate temp dir outside the build tree for the
 	# Unix-domain socket, to avoid file name length issues.
-	return File::Temp::tempdir(CLEANUP => 1);
+	return File::Temp::tempdir(CLEANUP => 0);
 }
 
 # Initialize a new cluster for testing.
@@ -217,6 +221,13 @@ END
 		system_log('pg_ctl', '-D', $test_server_datadir, '-m',
 		  'immediate', 'stop');
 	}
+
+	# Cleanup any temporary directory created
+	return if (!$temp_cleanup);
+	foreach my $temp_dir (@temp_dirs)
+	{
+		rmtree($temp_dir);
+	}
 }
 
 sub psql
@@ -278,14 +289,16 @@ sub command_ok
 {
 	my ($cmd, $test_name) = @_;
 	my $result = run_log($cmd);
-	ok($result, $test_name);
+	my $err_num = ok($result, $test_name);
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub command_fails
 {
 	my ($cmd, $test_name) = @_;
 	my $result = run_log($cmd);
-	ok(!$result, $test_name);
+	my $err_num = ok(!$result, $test_name);
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub command_exit_is
@@ -304,7 +317,8 @@ sub command_exit_is
 	# that, use $h->full_result on Windows instead.
 	my $result = ($Config{osname} eq "MSWin32") ?
 		($h->full_results)[0] : $h->result(0);
-	is($result, $expected, $test_name);
+	my $err_num = is($result, $expected, $test_name);
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub program_help_ok
@@ -313,9 +327,13 @@ sub program_help_ok
 	my ($stdout, $stderr);
 	print("# Running: $cmd --help\n");
 	my $result = run [ $cmd, '--help' ], '>', \$stdout, '2>', \$stderr;
-	ok($result, "$cmd --help exit code 0");
-	isnt($stdout, '', "$cmd --help goes to stdout");
-	is($stderr, '', "$cmd --help nothing to stderr");
+	my $err_num;
+	$err_num = ok($result, "$cmd --help exit code 0");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = isnt($stdout, '', "$cmd --help goes to stdout");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = is($stderr, '', "$cmd --help nothing to stderr");
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub program_version_ok
@@ -324,9 +342,13 @@ sub program_version_ok
 	my ($stdout, $stderr);
 	print("# Running: $cmd --version\n");
 	my $result = run [ $cmd, '--version' ], '>', \$stdout, '2>', \$stderr;
-	ok($result, "$cmd --version exit code 0");
-	isnt($stdout, '', "$cmd --version goes to stdout");
-	is($stderr, '', "$cmd --version nothing to stderr");
+	my $err_num;
+	$err_num = ok($result, "$cmd --version exit code 0");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = isnt($stdout, '', "$cmd --version goes to stdout");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = is($stderr, '', "$cmd --version nothing to stderr");
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub program_options_handling_ok
@@ -335,20 +357,27 @@ sub program_options_handling_ok
 	my ($stdout, $stderr);
 	print("# Running: $cmd --not-a-valid-option\n");
 	my $result = run [ $cmd, '--not-a-valid-option' ], '>', \$stdout, '2>',
-	  \$stderr;
-	ok(!$result, "$cmd with invalid option nonzero exit code");
-	isnt($stderr, '', "$cmd with invalid option prints error message");
+	\$stderr;
+	my $err_num;
+	$err_num = ok(!$result, "$cmd with invalid option nonzero exit code");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = isnt($stderr, '', "$cmd with invalid option prints error message");
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub command_like
 {
 	my ($cmd, $expected_stdout, $test_name) = @_;
 	my ($stdout, $stderr);
+	my $err_num;
 	print("# Running: " . join(" ", @{$cmd}) . "\n");
 	my $result = run $cmd, '>', \$stdout, '2>', \$stderr;
-	ok($result, "@$cmd exit code 0");
+	$err_num = ok($result, "@$cmd exit code 0");
+	$temp_cleanup = 0 if (!$err_num);
 	is($stderr, '', "@$cmd no stderr");
-	like($stdout, $expected_stdout, "$test_name: matches");
+	$temp_cleanup = 0 if (!$err_num);
+	$err_num = like($stdout, $expected_stdout, "$test_name: matches");
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 sub issues_sql_like
@@ -356,9 +385,12 @@ sub issues_sql_like
 	my ($cmd, $expected_sql, $test_name) = @_;
 	truncate $test_server_logfile, 0;
 	my $result = run_log($cmd);
-	ok($result, "@$cmd exit code 0");
+	my $err_num;
+	$err_num = ok($result, "@$cmd exit code 0");
+	$temp_cleanup = 0 if (!$err_num);
 	my $log = slurp_file($test_server_logfile);
-	like($log, $expected_sql, "$test_name: SQL found in server log");
+	$err_num = like($log, $expected_sql, "$test_name: SQL found in server log");
+	$temp_cleanup = 0 if (!$err_num);
 }
 
 1;
